# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path


def test_snapshot(fuse_graph_mountpoint: Path, example_snapshot: str):
    root = fuse_graph_mountpoint / "archive" / example_snapshot
    assert root.is_dir()
    subpaths = list(root.iterdir())
    assert len(subpaths) == 1
    assert subpaths[0].name == "refs"
    assert subpaths[0].is_dir()
    refs = subpaths[0]

    subfolders = set(d.name for d in refs.iterdir())
    assert subfolders == {"heads", "tags"}
    assert (refs / "heads").is_dir()
    assert (refs / "heads" / "master").is_symlink()
    assert (refs / "tags").is_dir()
    assert (refs / "tags" / "v1.0").is_symlink()



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information
from pathlib import Path

import pytest


def test_directory(fuse_graph_mountpoint: Path, example_directory: str):
    with pytest.raises(FileNotFoundError):
        # this should not crash the process
        void = fuse_graph_mountpoint / "archive" / example_directory / "not-there"
        void.stat()

    root = fuse_graph_mountpoint / "archive" / example_directory
    assert root.is_dir()

    for item in root.iterdir():
        if item.is_dir():
            # in the test graph, that subfolder only contains two files
            files = list(item.iterdir())
            assert len(files) == 2
            for file in files:
                assert file.is_file()
        elif item.is_file():
            assert item.stat().st_size > 0

        else:
            raise AssertionError(
                f"{item} is not a file or directory, that's unexpected in the test graph"
            )



# Copyright (C) 2020-2024  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
import errno
import logging
import os
from pathlib import Path
from shutil import rmtree, which
from subprocess import run
from tempfile import mkdtemp
from threading import Thread
import time
from typing import Any, Dict, List

import pyfuse3
import pyfuse3.asyncio

from swh.fuse import LOGGER_NAME
from swh.fuse.backends import ContentBackend, GraphBackend
from swh.fuse.cache import FuseCache
from swh.fuse.cli import load_config
from swh.fuse.fs.entry import FuseDirEntry, FuseEntry, FuseFileEntry, FuseSymlinkEntry
from swh.fuse.fs.mountpoint import Root
from swh.model.swhids import CoreSWHID, ObjectType


class Fuse(pyfuse3.Operations):
    """
    Software Heritage Filesystem in Userspace (FUSE).

    Locally mount parts of the archive and navigate it as a virtual file system.

    This class ties together ``pyfuse3`` and the configured cache and back-end.
    """

    def __init__(
        self,
        cache: FuseCache,
        conf: Dict[str, Any],
        graph_backend: GraphBackend,
        obj_backend: ContentBackend,
    ):
        super(Fuse, self).__init__()

        self._next_inode: pyfuse3.InodeT = pyfuse3.ROOT_INODE
        self._inode2entry: Dict[pyfuse3.InodeT, FuseEntry] = {}

        self.root = Root(fuse=self)
        self.conf = conf
        self.logger = logging.getLogger(LOGGER_NAME)

        self.time_ns: int = time.time_ns()  # start time, used as timestamp
        self.gid = os.getgid()
        self.uid = os.getuid()

        self.graph_backend = graph_backend
        self.obj_backend = obj_backend
        self.cache = cache

    def shutdown(self) -> None:
        self.graph_backend.shutdown()
        self.obj_backend.shutdown()

    def _alloc_inode(self, entry: FuseEntry) -> int:
        """Return a unique inode integer for a given entry"""

        inode = self._next_inode
        self._next_inode = pyfuse3.InodeT(inode + 1)
        self._inode2entry[inode] = entry

        return inode

    def _remove_inode(self, inode: pyfuse3.InodeT) -> None:
        try:
            del self._inode2entry[inode]
        except KeyError:
            pass

        try:
            pyfuse3.invalidate_inode(inode)
        except FileNotFoundError:
            pass

    def inode2entry(self, inode: pyfuse3.InodeT) -> FuseEntry:
        """Return the entry matching a given inode"""

        try:
            return self._inode2entry[inode]
        except KeyError:
            raise pyfuse3.FUSEError(errno.ENOENT)

    async def get_metadata(self, swhid: CoreSWHID) -> Any:
        """Retrieve metadata for a given SWHID using Software Heritage API"""

        cache = await self.cache.metadata.get(swhid)
        if cache is not None:
            return cache

        metadata = await self.graph_backend.get_metadata(swhid)
        await self.cache.metadata.set(swhid, metadata)
        # Retrieve it from cache so it is correctly typed
        return await self.cache.metadata.get(swhid)

    async def get_blob(self, swhid: CoreSWHID) -> bytes:
        """Retrieve the blob bytes for a given content SWHID using Software
        Heritage API"""

        if swhid.object_type != ObjectType.CONTENT:
            raise pyfuse3.FUSEError(errno.EINVAL)

        cache = await self.cache.blob.get(swhid)
        if cache is not None:
            self.logger.debug("Found blob %s in cache", swhid)
            return cache

        blob = await self.obj_backend.get_blob(swhid)
        await self.cache.blob.set(swhid, blob)
        return blob

    async def get_history(self, swhid: CoreSWHID) -> List[CoreSWHID]:
        """Retrieve a revision's history using Software Heritage Graph API"""

        if swhid.object_type != ObjectType.REVISION:
            raise pyfuse3.FUSEError(errno.EINVAL)

        cache = await self.cache.history.get(swhid)
        if cache is not None:
            self.logger.debug(
                "Found history of %s in cache (%d ancestors)", swhid, len(cache)
            )
            return cache

        history = await self.graph_backend.get_history(swhid)
        await self.cache.history.set(history)
        # Retrieve it from cache so it is correctly typed
        res = await self.cache.history.get(swhid)
        return res

    async def get_visits(self, url_encoded: str) -> List[Dict[str, Any]]:
        """Retrieve origin visits given an encoded-URL using Software Heritage API"""

        cache = await self.cache.metadata.get_visits(url_encoded)
        if cache is not None:
            self.logger.debug(
                "Found %d visits for origin '%s' in cache",
                len(cache),
                url_encoded,
            )
            return cache

        visits = await self.graph_backend.get_visits(url_encoded)
        await self.cache.metadata.set_visits(url_encoded, visits)
        # Retrieve it from cache so it is correctly typed
        res = await self.cache.metadata.get_visits(url_encoded)
        return res

    async def get_attrs(self, entry: FuseEntry) -> pyfuse3.EntryAttributes:
        """Return entry attributes"""

        attrs = pyfuse3.EntryAttributes()
        attrs.st_size = 0
        attrs.st_atime_ns = self.time_ns
        attrs.st_ctime_ns = self.time_ns
        attrs.st_mtime_ns = self.time_ns
        attrs.st_gid = self.gid
        attrs.st_uid = self.uid
        attrs.st_ino = pyfuse3.InodeT(entry.inode)
        attrs.st_mode = pyfuse3.ModeT(entry.mode)
        attrs.st_size = await entry.size()
        return attrs

    async def getattr(
        self, inode: pyfuse3.InodeT, _ctx: pyfuse3.RequestContext
    ) -> pyfuse3.EntryAttributes:
        """Get attributes for a given inode"""

        entry = self.inode2entry(inode)
        return await self.get_attrs(entry)

    async def opendir(
        self, inode: pyfuse3.InodeT, _ctx: pyfuse3.RequestContext
    ) -> pyfuse3.FileHandleT:
        """Open a directory referred by a given inode"""

        # Reuse inode as directory handle
        self.logger.debug("opendir(inode=%d)", inode)
        return pyfuse3.FileHandleT(inode)

    async def readdir(
        self, fh: pyfuse3.FileHandleT, offset: int, token: pyfuse3.ReaddirToken
    ) -> None:
        """Read entries in an open directory"""

        # opendir() uses inode as directory handle
        inode = pyfuse3.InodeT(fh)
        direntry = self.inode2entry(inode)
        self.logger.debug(
            "readdir(dirname=%s, fh=%d, offset=%d)", direntry.name, fh, offset
        )
        assert isinstance(direntry, FuseDirEntry)

        next_id = offset + 1
        try:
            async for entry in direntry.get_entries(offset):
                name = pyfuse3.FileNameT(os.fsencode(entry.name))
                attrs = await self.get_attrs(entry)
                if not pyfuse3.readdir_reply(token, name, attrs, next_id):
                    break

                next_id += 1
                self._inode2entry[attrs.st_ino] = entry
        except Exception as err:
            self.logger.exception("Cannot readdir: %s", err)
            raise pyfuse3.FUSEError(errno.ENOENT)

    async def open(
        self, inode: pyfuse3.InodeT, _flags: int, _ctx: pyfuse3.RequestContext
    ) -> pyfuse3.FileInfo:
        """Open an inode and return a unique file handle"""

        # Reuse inode as file handle
        self.logger.debug("open(inode=%d)", inode)
        entry = self.inode2entry(inode)
        return pyfuse3.FileInfo(fh=pyfuse3.FileHandleT(inode), **entry.file_info_attrs)

    async def read(self, fh: pyfuse3.FileHandleT, offset: int, length: int) -> bytes:
        """Read `length` bytes from file handle `fh` at position `offset`"""

        # open() uses inode as file handle
        inode = pyfuse3.InodeT(fh)

        entry = self.inode2entry(inode)
        self.logger.debug(
            "read(name=%s, fh=%d, offset=%d, length=%d)", entry.name, fh, offset, length
        )
        assert isinstance(entry, FuseFileEntry)

        try:
            data = await entry.get_content()
            return data[offset : offset + length]
        except Exception as err:
            self.logger.exception("Cannot read: %s", err)
            raise pyfuse3.FUSEError(errno.ENOENT)

    async def lookup(
        self,
        parent_inode: pyfuse3.InodeT,
        name: pyfuse3.FileNameT,
        _ctx: pyfuse3.RequestContext,
    ) -> pyfuse3.EntryAttributes:
        """Look up a directory entry by name and get its attributes"""

        decoded_name = os.fsdecode(name)
        parent_entry = self.inode2entry(parent_inode)
        self.logger.debug(
            "lookup(parent_name=%s, parent_inode=%d, name=%s)",
            parent_entry.name,
            parent_inode,
            decoded_name,
        )
        assert isinstance(parent_entry, FuseDirEntry)

        try:
            if parent_entry.validate_entry(decoded_name):
                lookup_entry = await parent_entry.lookup(decoded_name)
                if lookup_entry:
                    return await self.get_attrs(lookup_entry)
        except Exception as err:
            self.logger.exception("Cannot lookup: %s", err)

        raise pyfuse3.FUSEError(errno.ENOENT)

    async def readlink(
        self, inode: pyfuse3.InodeT, _ctx: pyfuse3.RequestContext
    ) -> pyfuse3.FileNameT:
        entry = self.inode2entry(inode)
        self.logger.debug("readlink(name=%s, inode=%d)", entry.name, inode)
        assert isinstance(entry, FuseSymlinkEntry)
        return pyfuse3.FileNameT(os.fsencode(entry.get_target()))

    async def unlink(
        self,
        parent_inode: pyfuse3.InodeT,
        name: pyfuse3.FileNameT,
        _ctx: pyfuse3.RequestContext,
    ) -> None:
        """Remove a file"""

        decoded_name = os.fsdecode(name)
        parent_entry = self.inode2entry(parent_inode)
        self.logger.debug(
            "unlink(parent_name=%s, parent_inode=%d, name=%s)",
            parent_entry.name,
            parent_inode,
            decoded_name,
        )

        try:
            await parent_entry.unlink(decoded_name)
        except Exception as err:
            self.logger.exception("Cannot unlink: %s", err)
            raise pyfuse3.FUSEError(errno.ENOENT)

    async def getxattr(
        self,
        inode: pyfuse3.InodeT,
        name: pyfuse3.XAttrNameT,
        _ctx: pyfuse3.RequestContext,
    ) -> bytes:
        """
        This allows someone to get the extend attribute "user.swhid" on entities that
        have one (this is mostly useful when traversing source trees).
        The attribute value is a SWHID string.
        """
        if name == b"user.swhid":
            entry = self.inode2entry(inode)
            try:
                return str(entry.swhid).encode()  # type: ignore
            except AttributeError:
                pass
        raise pyfuse3.FUSEError(errno.ENOSYS)


def graph_backend_factory(conf: Dict[str, Any]) -> GraphBackend:
    if "graph" in conf:
        from swh.fuse.backends.compressed import CompressedGraphBackend

        return CompressedGraphBackend(conf)
    else:
        from swh.fuse.backends.web_api import WebApiBackend

        return WebApiBackend(conf)


def obj_backend_factory(conf: Dict[str, Any]) -> ContentBackend:
    if "content" in conf and conf["content"]:
        from swh.fuse.backends.objstorage import ObjStorageBackend

        return ObjStorageBackend(conf)
    else:
        from swh.fuse.backends.web_api import WebApiBackend

        return WebApiBackend(conf)


class SwhFsTmpMount:
    """
    This context manager will mount the Software Heritage archive on a temporary
    folder, in a separate thread running its own asyncio event loop. It returns a
    `Path <https://docs.python.org/3/library/pathlib.html>`_ object pointing to the
    mountpoint, that will be deleted when exiting the context.
    Note that the main thread will likely
    wait a bit before entering the context, until the mountpoint appears.

    Example:

    ::

        with SwhFsTmpMount() as mountpoint:
            swhid = "swh:1:cnt:c839dea9e8e6f0528b468214348fee8669b305b2"
            hello_world_path = mountpoint / "archive" / swhid
            print(open(hello_world_path).read())

    SwhFS will be configured as if launched via the `swh fs mount` command,
    so please set the ``SWH_CONFIG_FILE`` environment variable pointing to the relevant
    configuration file. The ``config`` parameter is intended for unit tests.

    See also :ref:`swh-fuse-parallelization`.
    """

    def __init__(self, config=None):
        if config is None:
            self.config = load_config()
        else:
            self.config = config
        self.mountpoint = Path(mkdtemp())
        self.loop = asyncio.get_event_loop_policy().get_event_loop()
        self.swhfuse = self.loop.create_task(main([], self.mountpoint, self.config))
        self.thread = Thread(target=self.loop.run_until_complete, args=(self.swhfuse,))

    def __enter__(self) -> Path:
        self.thread.start()
        for i in range(30000):
            if (self.mountpoint / "archive").exists():
                break
            else:
                time.sleep(0.01)
        else:
            raise RuntimeError("Mountpoint failed to appear after 5 minutes, aborting.")
        return self.mountpoint

    def __exit__(self, exc_type, exc_value, traceback):
        if (self.mountpoint / "archive").exists():
            run([which("fusermount3"), "-u", str(self.mountpoint)])
        rmtree(self.mountpoint, ignore_errors=True)
        return False


async def main(
    swhids: List[CoreSWHID],
    root_path: Path,
    conf: Dict[str, Any],
    content_backend: ContentBackend | None = None,
) -> None:
    """``swh-fuse`` CLI entry-point"""

    # Use pyfuse3 asyncio layer to match the rest of Software Heritage codebase
    pyfuse3.asyncio.enable()

    async with FuseCache(conf["cache"]) as cache:
        if content_backend is None:
            content_backend = obj_backend_factory(conf)
        fs = Fuse(cache, conf, graph_backend_factory(conf), content_backend)

        # Initially populate the cache
        for swhid in swhids:
            try:
                await fs.get_metadata(swhid)
            except Exception as err:
                fs.logger.exception("Cannot prefetch object %s: %s", swhid, err)

        fuse_options = set(pyfuse3.default_options)
        fuse_options.add("fsname=swhfs")

        try:
            pyfuse3.init(fs, str(root_path), fuse_options)
            await pyfuse3.main()
        except Exception as err:
            fs.logger.error("Error running FUSE: %s", err)
        finally:
            fs.shutdown()
            pyfuse3.close(unmount=True)



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
FUSE backends
-------------

These classes are abstraction layers to the archive's graph and object storage.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, List, Tuple

from swh.model.swhids import CoreSWHID


class GraphBackend(ABC):
    """
    Methods required by the FUSE logic to create its folders layout.
    Those methods must return dicts compliant with the Web API, to ease their use and
    caching in the :py:class:`swh.fuse.fuse.Fuse` class.
    Fields listed in method specifications below are the minimal required fields.
    At worst, set them to ``None`` (because this is better supported by
    ``swh-web-client``'s ``typify_json``).
    Additional fields will be included in the various `meta.json` files proposed
    by this virtual file system.
    """

    @abstractmethod
    async def get_metadata(self, swhid: CoreSWHID) -> Dict | List:
        """
        Entries in the returned ``dict`` depend on the ``swhid`` type.

        For ``cnt``, return a dict containing at least ``length`` (int).

        For ``dir``, return a list of entries like

        .. code:: python

            {
                "dir_id": "dir's hash",
                "name": "dir item name",
                "type": "str=dir|file|rev",  # yes, `file`...
                "target": "target hash",
                "perms": "permissions [int]",
                # if target is a cnt:
                "length": "file size [int]",
            }

        For ``rev``, return a dict containing at least ``directory`` (SWHID),
        ``parents`` (list of objects containing at least ``id`` (hash)),
        ``date`` (str, isoformat), ``committer_date`` (str, isoformat),
        ``id`` (object's hash).

        For ``rel``, return a dict containing at least ``target`` (SWHID),
        ``target_type``,  ``date`` (str, isoformat), ``id`` (object's hash).

        For ``snp``, return a dict where each key is a branch name,
        each value is a dict containing at least ``target`` (hash),
        ``target_type`` (content, directory, revision, release, snapshot or alias).
        """

    @abstractmethod
    async def get_history(self, swhid: CoreSWHID) -> List[Tuple[str, str]]:
        """
        Return a list of tuples ``(swhid, revision swhid)``
        """

    @abstractmethod
    async def get_visits(self, url_encoded: str) -> List[Dict[str, Any]]:
        """
        Return a list of objects

        Each object should contain fields ``date`` (ISO), ``origin`` (str),
        ``snapshot`` (SWHID's hash as str)
        """

    def shutdown(self) -> None:
        """
        Called when the FUSE object is destroyed.
        """


class ContentBackend(ABC):
    """
    Methods required by the FUSE logic to provide files' contents.
    """

    @abstractmethod
    async def get_blob(self, swhid: CoreSWHID) -> bytes:
        """
        Fetch the content of a ``cnt`` object.
        """

    def shutdown(self) -> None:
        """
        Called when the FUSE object is destroyed.
        """



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

from pathlib import Path


def test_origin(fuse_graph_mountpoint: Path, example_origin: str):
    root = fuse_graph_mountpoint / "origin" / example_origin
    assert root.is_dir()
    entries = list(root.iterdir())
    for d in entries:
        assert d.is_dir()
    if not entries:
        raise AssertionError(f"No directories in {root}")
    assert (d / "meta.json").is_file()
    assert (d / "snapshot").is_symlink()



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import asyncio
from functools import partial
import logging
from typing import Any, Dict, List, Tuple
from urllib.parse import unquote_plus

from requests import HTTPError

from swh.fuse import LOGGER_NAME
from swh.fuse.backends import ContentBackend, GraphBackend
from swh.model.swhids import CoreSWHID
from swh.web.client.client import WebAPIClient


class WebApiBackend(GraphBackend, ContentBackend):
    """
    A Backend querying everything via Software Heritage's public API.

    This is simpler to configure and deploy, but expect long response times.
    """

    def __init__(self, conf: Dict):
        """
        Only needs the ``web-api`` key of ``conf``, searching for ``url`` and maybe
        ``auth-token`` keys.
        """
        self.web_api = WebAPIClient(
            conf["web-api"]["url"], conf["web-api"]["auth-token"]
        )
        self.logger = logging.getLogger(LOGGER_NAME)

    async def get_metadata(self, swhid: CoreSWHID) -> Dict | List:
        try:
            self.logger.debug(f"Fetching metadata via Web API for {swhid}")
            loop = asyncio.get_event_loop()
            metadata = await loop.run_in_executor(None, self.web_api.get, swhid, False)
            return metadata
        except HTTPError as err:
            self.logger.error("Cannot fetch metadata for object %s: %s", swhid, err)
            raise

    async def get_blob(self, swhid) -> bytes:
        try:
            self.logger.debug("Retrieving blob %s via web API...", swhid)
            loop = asyncio.get_event_loop()
            resp = await loop.run_in_executor(None, self.web_api.content_raw, swhid)
            blob = b"".join(list(resp))
            return blob
        except HTTPError as err:
            self.logger.error("Cannot fetch blob for object %s: %s", swhid, err)
            raise

    async def get_history(self, swhid: CoreSWHID) -> List[Tuple[str, str]]:
        try:
            # Use the swh-graph API to retrieve the full history very fast
            self.logger.debug("Retrieving history of %s via graph API...", swhid)
            call = f"graph/visit/edges/{swhid}?edges=rev:rev"
            loop = asyncio.get_event_loop()
            request = await loop.run_in_executor(None, self.web_api._call, call)
            history = request.text.strip()
            if history:
                edges = []
                for edge in history.split("\n"):
                    split = edge.split(" ")
                    if len(split) == 2:
                        edges.append((split[0], split[1]))
                return edges
            else:
                return []
        except HTTPError as err:
            self.logger.error("Cannot fetch history for object %s: %s", swhid, err)
            # Ignore exception since swh-graph does not necessarily contain the
            # most recent artifacts from the archive. Computing the full history
            # from the Web API is too computationally intensive so simply return
            # an empty list.
            return []

    async def get_visits(self, url_encoded: str) -> List[Dict[str, Any]]:
        try:
            self.logger.debug(
                "Retrieving visits for origin '%s' via web API...", url_encoded
            )
            loop = asyncio.get_event_loop()
            # Web API only takes non-encoded URL
            url = unquote_plus(url_encoded)

            origin_exists = await loop.run_in_executor(
                None, self.web_api.origin_exists, url
            )
            if not origin_exists:
                raise ValueError("origin does not exist")

            visits_it = await loop.run_in_executor(
                None, partial(self.web_api.visits, url, typify=False)
            )
            visits = list(visits_it)
            return visits
        except (ValueError, HTTPError) as err:
            self.logger.error(
                "Cannot fetch visits for origin '%s': %s", url_encoded, err
            )
            raise



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

"""
Fixtures for the GraphBackend-ed swh-fuse
-----------------------------------------

Those rely on the test graph in `swh-graph`.
Data expected in the test graph is provided here as fixtures and will have to be kept
in sync with swh/graph/tests/grpc/test_getnode.py
"""

import asyncio
import logging
import os
from pathlib import Path
from tempfile import TemporaryDirectory
import threading
import time
from typing import Generator
from urllib.parse import quote_plus

from click.testing import CliRunner
import pytest

from swh.fuse import LOGGER_NAME, fuse
from swh.fuse.backends.objstorage import ObjStorageBackend
import swh.fuse.cli as cli
from swh.graph.pytest_plugin import *  # noqa ; this provides the graph_grpc_server fixture
from swh.model.model import Content
from swh.objstorage.interface import objid_from_dict


@pytest.fixture(scope="module")
def fuse_graph_mountpoint(
    graph_grpc_server, example_content: Content
) -> Generator[Path, None, None]:
    with TemporaryDirectory(suffix=".swh-fuse-test") as tmpdir:
        content_backend = ObjStorageBackend(
            {
                "content": {
                    "storage": {
                        "cls": "memory",
                        "journal_writer": {
                            "cls": "memory",
                        },
                    },
                    "objstorage": {
                        "cls": "memory",
                    },
                }
            }
        )
        if content_backend.storage is not None:
            content_backend.storage.content_add([example_content])
        if content_backend.objstorage:
            if example_content.data:
                content_backend.objstorage.add(
                    example_content.data, objid_from_dict(example_content.hashes())
                )
        else:
            raise AssertionError("ObjStorageBackend should yield an objstorage")

        mountpoint = Path(tmpdir)
        config = {
            "graph": {
                "grpc-url": graph_grpc_server,
            },
            "cache": {
                "metadata": {"in-memory": True},
                "blob": {"bypass": True},
                "direntry": {"maxram": "10%"},
            },
            "json-indent": None,
        }

        def mount():
            loop = asyncio.new_event_loop()
            logging.getLogger(LOGGER_NAME).setLevel(logging.DEBUG)
            try:
                loop.run_until_complete(
                    fuse.main([], mountpoint, config, content_backend)
                )
            finally:
                if loop.is_running():
                    loop.stop()
                loop.close()

        mount_thread = threading.Thread(target=mount)
        mount_thread.start()

        for i in range(30):
            try:
                root = os.listdir(tmpdir)
                if root:
                    break
            except FileNotFoundError:
                pass
            time.sleep(0.1)
        else:
            raise FileNotFoundError(f"Could not mount FUSE in {tmpdir}")

        try:
            yield mountpoint
        finally:
            CliRunner().invoke(cli.umount, [tmpdir])


@pytest.fixture()
def example_origin() -> str:
    "Adapted from swh.graph.tests.test_http_client.TEST_ORIGIN_ID"
    return quote_plus("https://example.com/swh/graph")


@pytest.fixture()
def example_snapshot() -> str:
    return "swh:1:snp:0000000000000000000000000000000000000020"


@pytest.fixture()
def example_revision() -> str:
    return "swh:1:rev:0000000000000000000000000000000000000009"


@pytest.fixture()
def example_release() -> str:
    return "swh:1:rel:0000000000000000000000000000000000000010"


@pytest.fixture()
def example_directory() -> str:
    return "swh:1:dir:0000000000000000000000000000000000000008"


@pytest.fixture()
def example_content_swhid() -> str:
    return "swh:1:cnt:0000000000000000000000000000000000000001"


@pytest.fixture(scope="module")
def example_content() -> Content:
    h = bytes.fromhex("0000000000000000000000000000000000000001")
    return Content(
        sha1_git=h,
        blake2s256=h,
        sha256=h,
        sha1=h,
        data=b"Hello world",
        length=11,
    )



# Copyright (C) 2025  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import json
from pathlib import Path


def test_release(fuse_graph_mountpoint: Path, example_release: str):
    root = fuse_graph_mountpoint / "archive" / example_release
    assert root.is_dir()

    meta = root / "meta.json"
    assert meta.is_symlink()
    with meta.open() as f:
        parsed = json.loads(f.read())
        assert parsed["id"] in example_release
        assert "name" in parsed
        assert "author" in parsed
        assert "message" in parsed
        assert parsed["date"] == "2009-02-14T01:31:30+02:00"

    assert (root / "root").is_symlink()
    assert (root / "target").is_symlink()

    with (root / "target_type").open() as f:
        assert f.read() == "revision\n"



#!/usr/bin/env python

"""
Launch WORKERS_N sub-process in parallel, to count *.py files in a dir list (SWH dir
20-bytes IDs, concatenated in concatenated_dir_ids.bin).
For each dir, this mounts one instance of swh-fuse over a temporary mountpoint.

If you are not authorized to FUSE-mount on your machine, wrap this script in a
namespace:

    unshare --pid --kill-child --user --map-root-user --mount ./parallel_processor.py 4
"""

from concurrent.futures import ProcessPoolExecutor
from datetime import datetime
from pathlib import Path
from time import perf_counter

import click

from swh.fuse.fuse import SwhFsTmpMount

DIRLIST = "concatenated_dir_ids.bin"


def count_py_worker(swhdir: Path) -> int:
    nbfiles = 0
    with SwhFsTmpMount() as mountpoint:
        path = mountpoint / swhdir
        for f in path.glob("**/*.py"):
            nbfiles += 1
    print(datetime.now().isoformat(), f"Found {nbfiles} .py files in {swhdir}")
    return nbfiles


def gen_paths():
    """
    Generates paths to each directory SWHID we'd like to analyze, relative to an
    SwhFS mountpoint.
    """
    with open(DIRLIST, "rb") as dirfile:
        while dirid := dirfile.read(20):
            yield Path("archive") / Path("swh:1:dir:" + dirid.hex())


@click.command(help=__doc__)
@click.argument("workers_n", type=int)
def main(workers_n: int):
    global_start = perf_counter()
    print(datetime.now().isoformat(), "starting")

    with SwhFsTmpMount():
        # some python imports are made only when mounting, so we mount once to
        # indirectly trigger those imports. This ensure that processes forked next
        # by ProcessPoolExecutor will be fully loaded.
        pass

    total_py = 0
    with ProcessPoolExecutor(max_workers=workers_n) as executor:
        results = executor.map(count_py_worker, gen_paths())
        for counter in results:
            total_py += counter

    print(
        datetime.now().isoformat(),
        f"Found {total_py} .py files in {(perf_counter() - global_start):.2f} seconds",
    )


if __name__ == "__main__":
    main()
